@@ -1,7 +1,7 @@
 """
 A collection of functions for integrating with the Archimedes API
 """
-
+import concurrent.futures
 import json
 import os
 from functools import partial
@@ -68,22 +68,15 @@ def get_intraday_trades(
         DataFrame with all the time series data
 
     Raises:
-        HTTPError: If an http error occurs when requesting the API.
+        HTTPError: If an HTTP error occurs when requesting the API.
         NoneAuth: If the user is unauthorized or if the authorization has expired.
     """
 
     if isinstance(price_areas, str):
         price_areas = [price_areas]
 
-    if start is None:
-        start = archimedes.ArchimedesConstants.DATE_LOW
-    else:
-        start = pd.to_datetime(start)
-
-    if end is None:
-        end = archimedes.ArchimedesConstants.DATE_HIGH
-    else:
-        end = pd.to_datetime(end)
+    start = pd.to_datetime(start) if start else archimedes.ArchimedesConstants.DATE_LOW
+    end = pd.to_datetime(end) if end else archimedes.ArchimedesConstants.DATE_HIGH
 
     queries = [
         {
@@ -102,14 +95,19 @@ def get_intraday_trades(
 
     base_url = get_api_base_url_v2()
     observation_data = []
-    for query in queries:
-        observation_data.extend(
-            _make_api_request(
+    with concurrent.futures.ThreadPoolExecutor() as executor:
+        future_requests = [
+            executor.submit(
+                _make_api_request,
                 f"{base_url}/observation_json/get",
                 access_token=access_token,
                 params=query,
             )
-        )
+            for query in queries
+        ]
+
+        for f in future_requests:
+            observation_data.extend(f.result())
 
     if not len(observation_data):
         return pd.DataFrame(
@@ -197,25 +195,21 @@ def get(
         DataFrame with all the time series data
 
     Raises:
-        HTTPError: If an http error occurs when requesting the API.
+        HTTPError: If an HTTP error occurs when requesting the API.
         NoneAuth: If the user is unauthorized or if the authorization has expired.
     """
 
+    if not series_ids:
+        return pd.DataFrame()
+
     if isinstance(series_ids, str):
         series_ids = [series_ids]
 
     if isinstance(price_areas, str):
         price_areas = [price_areas]
 
-    if start is None:
-        start = archimedes.ArchimedesConstants.DATE_LOW
-    else:
-        start = pd.to_datetime(start)
-
-    if end is None:
-        end = archimedes.ArchimedesConstants.DATE_HIGH
-    else:
-        end = pd.to_datetime(end)
+    start = pd.to_datetime(start) if start else archimedes.ArchimedesConstants.DATE_LOW
+    end = pd.to_datetime(end) if end else archimedes.ArchimedesConstants.DATE_HIGH
 
     queries = [
         {
@@ -238,98 +232,124 @@ def get(
         ]
 
     df = None
-    for query in queries:
-        base_url = get_api_base_url_v2()
-        observations = _make_api_request(
-            f"{base_url}/data/get",
-            access_token=access_token,
-            params=query,
-        )
-        if len(observations):
+    base_url = get_api_base_url_v2()
+    with concurrent.futures.ThreadPoolExecutor() as executor:
+        future_requests = {
+            "observations": [],
+            "observations_json": [],
+        }
+        try:
+            for query in queries:
+                future_requests["observations"].append(
+                    executor.submit(
+                        _make_api_request,
+                        f"{base_url}/data/get",
+                        access_token=access_token,
+                        params=query,
+                    )
+                )
+                future_requests["observations_json"].append(
+                    executor.submit(
+                        _make_api_request,
+                        f"{base_url}/observation_json/get",
+                        access_token=access_token,
+                        params=query,
+                    )
+                )
+
+            api_results = {
+                "observations": [f.result() for f in future_requests["observations"]],
+                "observations_json": [
+                    f.result() for f in future_requests["observations_json"]
+                ],
+            }
+        except KeyboardInterrupt:
+            print("Cancelling requests")
+            cancelled_count = 0
+            total_count = 0
+            for _, fs in future_requests.items():
+                for f in fs:
+                    total_count += 1
+                    if f.cancel():
+                        cancelled_count += 1
+            print(
+                f"Cancelled {cancelled_count}/{total_count} requests. "
+                f"Waiting for {total_count - cancelled_count} requests to complete..."
+            )
+            exit()
+
+    for observation_type, observations_list in api_results.items():
+        for observations in observations_list:
+            if len(observations) == 0:
+                continue
+
             observations = pd.DataFrame.from_dict(observations)
-            observations = observations.sort_values(by=["from_dt", "version"])
-            observations = observations.pivot_table(
-                values="value",
-                columns=["series_id", "price_area"],
-                index="from_dt",
-                aggfunc="last",
-            )
-        else:
-            observations = pd.DataFrame()
 
-        observation_json = _make_api_request(
-            f"{base_url}/observation_json/get",
-            access_token=access_token,
-            params=query,
-        )
-        if len(observation_json):
-            observation_json = pd.DataFrame.from_dict(observation_json)
-
-            observation_json[
-                ["series_id1", "series_id2", "price_area"]
-            ] = observation_json["series_id"].str.split("/", 2, expand=True)
-            observation_json["series_id"] = (
-                observation_json["series_id1"] + "/" + observation_json["series_id2"]
-            )
-            observation_json = observation_json.drop(
-                ["series_id1", "series_id2"], axis=1
-            )
-            is_intraday_trades = observation_json["series_id"] == "NP/IntradayTrades"
-
-            def agg_func(x):
-                return list(x)
-
-            observation_data_intraday = observation_json[is_intraday_trades]
-            if len(observation_data_intraday):
-                observation_data_intraday = observation_data_intraday.sort_values(
-                    by=["from_dt"]
-                )
-                observation_data_intraday = observation_data_intraday.pivot_table(
-                    values="value",
-                    columns=["series_id", "price_area"],
-                    index="from_dt",
-                    aggfunc=agg_func,
-                )
-            else:
-                observation_data_intraday = pd.DataFrame()
-            observation_json = observation_json[~is_intraday_trades]
-            if len(observation_json):
-                observation_json = observation_json.sort_values(
-                    by=["from_dt", "version"]
-                ).pivot_table(
+            if observation_type == "observations":
+                observations = observations.sort_values(by=["from_dt", "version"])
+                observations = observations.pivot_table(
                     values="value",
                     columns=["series_id", "price_area"],
                     index="from_dt",
                     aggfunc="last",
                 )
-            else:
-                observation_json = pd.DataFrame()
+            elif observation_type == "observations_json":
+                observations[["series_id1", "series_id2", "price_area"]] = observations[
+                    "series_id"
+                ].str.split("/", 2, expand=True)
+                observations["series_id"] = (
+                    observations["series_id1"] + "/" + observations["series_id2"]
+                )
+                observations = observations.drop(["series_id1", "series_id2"], axis=1)
+                is_intraday_trades = observations["series_id"] == "NP/IntradayTrades"
 
-            observation_json = observation_json.merge(
-                observation_data_intraday,
-                left_index=True,
-                right_index=True,
-                how="outer",
-            )
+                def agg_func(x):
+                    return list(x)
 
-            observations = pd.merge(
-                observations,
-                observation_json,
-                left_index=True,
-                right_index=True,
-                how="outer",
+                observation_data_intraday = observations[is_intraday_trades]
+                if len(observation_data_intraday):
+                    observation_data_intraday = observation_data_intraday.sort_values(
+                        by=["from_dt"]
+                    )
+                    observation_data_intraday = observation_data_intraday.pivot_table(
+                        values="value",
+                        columns=["series_id", "price_area"],
+                        index="from_dt",
+                        aggfunc=agg_func,
+                    )
+                else:
+                    observation_data_intraday = pd.DataFrame()
+                observations = observations[~is_intraday_trades]
+                if len(observations):
+                    observations = observations.sort_values(
+                        by=["from_dt", "version"]
+                    ).pivot_table(
+                        values="value",
+                        columns=["series_id", "price_area"],
+                        index="from_dt",
+                        aggfunc="last",
+                    )
+                else:
+                    observations = pd.DataFrame()
+
+                observations = observations.merge(
+                    observation_data_intraday,
+                    left_index=True,
+                    right_index=True,
+                    how="outer",
+                )
+
+            df = (
+                observations
+                if df is None
+                else pd.merge(
+                    df,
+                    observations,
+                    left_index=True,
+                    right_index=True,
+                    how="outer",
+                )
             )
-        df = (
-            observations
-            if df is None
-            else pd.merge(
-                df,
-                observations,
-                left_index=True,
-                right_index=True,
-                how="outer",
-            )
-        )
 
     df.index = pd.to_datetime(df.index)
     return df
@@ -373,7 +393,7 @@ def get_latest(
         DataFrame with the latest time series data
 
     Raises:
-        HTTPError: If an http error occurs when requesting the API.
+        HTTPError: If an HTTP error occurs when requesting the API.
         NoneAuth: If the user is unauthorized or if the authorization has expired.
     """
     now_dt = pd.Timestamp.now(tz="utc")
@@ -469,18 +489,25 @@ def get_predictions(
 
     base_url = get_api_base_url_v2()
     data = []
-    for query in queries:
-        data.extend(
-            _make_api_request(
+    with concurrent.futures.ThreadPoolExecutor() as executor:
+        future_requests = [
+            executor.submit(
+                _make_api_request,
                 f"{base_url}/data/get_predictions",
                 access_token=access_token,
                 params=query,
             )
-        )
-    for item in data:
-        json_data = item.get("json_data")
-        item.update(json_data)
-        del item["json_data"]
+            for query in queries
+        ]
+
+        for f in future_requests:
+            for item in f.result():
+                if "json_data" in item:
+                    json_data = item.get("json_data")
+                    item.update(json_data)
+                    del item["json_data"]
+                data.append(item)
+
     df = pd.DataFrame.from_dict(data)
 
     date_fields = ["from_dt", "run_dt", "ref_dt"]
@@ -916,7 +943,7 @@ def forecast_get(
         DataFrame with all the time series data
 
     Raises:
-        HTTPError: If an http error occurs when requesting the API.
+        HTTPError: If an HTTP error occurs when requesting the API.
         NoneAuth: If the user is unauthorized or if the authorization has expired.
     """
 
@@ -926,15 +953,8 @@ def forecast_get(
     if isinstance(price_areas, str):
         price_areas = [price_areas]
 
-    if start is None:
-        start = archimedes.ArchimedesConstants.DATE_LOW
-    else:
-        start = pd.to_datetime(start)
-
-    if end is None:
-        end = archimedes.ArchimedesConstants.DATE_HIGH
-    else:
-        end = pd.to_datetime(end)
+    start = pd.to_datetime(start) if start else archimedes.ArchimedesConstants.DATE_LOW
+    end = pd.to_datetime(end) if end else archimedes.ArchimedesConstants.DATE_HIGH
 
     queries = [
         {
@@ -955,18 +975,23 @@ def forecast_get(
     ]
 
     base_url = get_api_base_url_v2()
-    df = []
-    for query in queries:
-        df.extend(
-            _make_api_request(
+    data = []
+    with concurrent.futures.ThreadPoolExecutor() as executor:
+        future_requests = [
+            executor.submit(
+                _make_api_request,
                 f"{base_url}/forecast/get",
                 access_token=access_token,
                 params=query,
             )
-        )
+            for query in queries
+        ]
 
-    if len(df):
-        df = pd.DataFrame.from_dict(df)
+        for f in future_requests:
+            data.extend(f.result())
+
+    if len(data):
+        df = pd.DataFrame.from_dict(data)
         df = df.sort_values(by=["from_dt", "ref_dt"])
     else:
         df = pd.DataFrame(
